
# Note: The following import statement is defined in order to resolve
# the conflict of the module names between 'os_ken.lib.ovs' and 'ovs'.
# The latter is the official Python package of Open vSwitch.
import ovs
